# -*- coding: utf-8 -*-
import termcolor

from . import __version__


class DTShellConstants:
    ROOT = "~/.dt-shell/"
    ENV_COMMANDS = "DTSHELL_COMMANDS"

    # DT1_TOKEN_ENV_VARIABLE = 'DUCKIETOWN_TOKEN'
    DT1_TOKEN_CONFIG_KEY = "token_dt1"
    CONFIG_DOCKER_USERNAME = "docker_username"
    CONFIG_DUCKIETOWN_VERSION = "duckietown_version"


ALLOWED_BRANCHES = ["daffy", "master19"]

DEBUG = False

CHECK_CMDS_UPDATE_EVERY_MINS = 5

DNAME = "Duckietown Shell"

INTRO = """

Welcome to the {Duckietown} ({version}).

Type "help" or "?" to list commands.

""".format(
    Duckietown=termcolor.colored(DNAME, "yellow", attrs=["bold"]), version=__version__
).lstrip()



import json
import os

import base58
import ecdsa
from ecdsa import BadSignatureError, SigningKey, VerifyingKey


class DuckietownToken:
    VERSION = "dt1"

    def __init__(self, payload, signature):
        self.payload = payload
        self.signature = signature

    def as_string(self):
        payload_58 = base58.b58encode(self.payload).decode("utf-8")
        signature_58 = base58.b58encode(self.signature).decode("utf-8")
        return "%s-%s-%s" % (DuckietownToken.VERSION, payload_58, signature_58)

    @staticmethod
    def from_string(s):
        p = s.split("-")
        if len(p) != 3:
            raise ValueError(p)
        if p[0] != DuckietownToken.VERSION:
            raise ValueError(p[0])
        payload_base58 = p[1]
        signature_base58 = p[2]
        payload = base58.b58decode(payload_base58)
        signature = base58.b58decode(signature_base58)
        return DuckietownToken(payload, signature)


private = "key1.pem"
public = "key1-pub.pem"
curve = ecdsa.NIST192p


def get_signing_key():
    if not os.path.exists(private):
        print("Creating private key %r" % private)
        sk0 = SigningKey.generate(curve=curve)
        with open(private, "w") as f:
            f.write(sk0.to_pem())

        vk = sk0.get_verifying_key()
        with open(public, "w") as f:
            f.write(vk.to_pem())
    pem = open(private).read()
    sk = SigningKey.from_pem(pem)
    return sk


def get_verify_key():
    key1 = """-----BEGIN PUBLIC KEY-----
MEkwEwYHKoZIzj0CAQYIKoZIzj0DAQEDMgAEQr/8RJmJZT+Bh1YMb1aqc2ao5teE
ixOeCMGTO79Dbvw5dGmHJLYyNPwnKkWayyJS
-----END PUBLIC KEY-----"""
    return VerifyingKey.from_pem(key1)


def create_signed_token(payload):
    sk = get_signing_key()

    def entropy(numbytes):
        s = b"duckietown is a place of relaxed introspection"
        return s[:numbytes]

    signature = sk.sign(payload, entropy=entropy)
    return DuckietownToken(payload, signature)


def verify_token(token):
    vk = get_verify_key()
    return vk.verify(token.signature, token.payload)


class InvalidToken(Exception):
    pass


def get_id_from_token(s):
    """
        Returns a numeric ID from the token, or raises InvalidToken.

    """
    try:
        token = DuckietownToken.from_string(s)
    except ValueError:
        msg = "Invalid token format %r." % s
        raise InvalidToken(msg)
    try:
        data = json.loads(token.payload)
        uid = data["uid"]
        return uid
    except ValueError:
        raise InvalidToken()


SAMPLE_TOKEN = (
    "dt1-9Hfd69b5ythetkCiNG12pKDrL987sLJT6KejWP2Eo5QQ"
    "-43dzqWFnWd8KBa1yev1g3UKnzVxZkkTbfWWn6of92V5Bf8qGV24rZHe6r7sueJNtWF"
)
SAMPLE_TOKEN_UID = -1
SAMPLE_TOKEN_EXP = "2018-10-20"


def tests_private():
    payload = json.dumps({"uid": SAMPLE_TOKEN_UID, "exp": SAMPLE_TOKEN_EXP})
    # generate a token
    token = create_signed_token(payload)
    s = token.as_string()
    print(s)
    assert s == SAMPLE_TOKEN
    token2 = token.from_string(s)

    assert verify_token(token2)


def test1():
    token = DuckietownToken.from_string(SAMPLE_TOKEN)
    assert verify_token(token)
    data = json.loads(token.payload)
    print(data)
    assert data["uid"] == SAMPLE_TOKEN_UID
    assert data["exp"] == SAMPLE_TOKEN_EXP

    seq = SAMPLE_TOKEN[6:8]
    msg_bad = SAMPLE_TOKEN.replace(seq, "XY")
    token = DuckietownToken.from_string(msg_bad)
    try:
        verify_token(token)
    except BadSignatureError:
        pass
    else:
        raise Exception(token)

    assert SAMPLE_TOKEN_UID == get_id_from_token(SAMPLE_TOKEN)

    try:
        get_id_from_token(msg_bad)
    except InvalidToken:
        pass
    else:
        raise Exception()


if __name__ == "__main__":
    if os.path.exists(private):
        tests_private()
    test1()



__all__ = ["InvalidEnvironment", "UserError", "CommandsLoadingException"]


class InvalidEnvironment(Exception):
    pass


class UserError(Exception):
    """ an error that will be briefly printed"""

    pass


class InvalidConfig(Exception):
    pass


class CommandsLoadingException(Exception):
    pass


class ConfigInvalid(Exception):
    pass


class ConfigNotPresent(Exception):
    pass


class CouldNotGetVersion(Exception):
    pass


class URLException(Exception):
    pass


class NoCacheAvailable(Exception):
    pass



# -*- coding: utf-8 -*-
import os
import sys
import time
import types
from cmd import Cmd
from dataclasses import dataclass
from os import remove, utime
from os.path import exists, isfile, join
from typing import Mapping, Optional, Sequence

from . import dtslogger
from .commands_ import (
    _get_commands,
    _init_commands,
    check_commands_outdated,
    InvalidRemote,
    update_commands,
)
from .config import (
    get_config_path,
    RepoInfo,
    RepoInfo_for_version,
    ShellConfig,
    write_shell_config,
)
from .constants import DEBUG, DTShellConstants, INTRO
from .dt_command_abs import DTCommandAbs
from .dt_command_placeholder import DTCommandPlaceholder
from .exceptions import CommandsLoadingException, UserError
from .logging import dts_print
from .version_check import check_if_outdated


@dataclass
class CommandsInfo:
    commands_path: str  # commands path
    leave_alone: bool  # whether to leave this alone (local)


def get_local_commands_info() -> CommandsInfo:
    # define commands_path
    config_path = get_config_path()
    V = DTShellConstants.ENV_COMMANDS
    if V in os.environ:
        commands_path = os.environ[V]

        if not os.path.exists(commands_path):
            msg = (
                "The path %s that you gave with the env. variable %s does not exist."
                % (commands_path, V)
            )
            raise Exception(msg)

        commands_path_leave_alone = True
        msg = "Using path %r as prescribed by env variable %s." % (commands_path, V)
        dtslogger.info(msg)
    else:
        commands_path = join(config_path, "commands-multi")
        commands_path_leave_alone = False
    return CommandsInfo(commands_path, commands_path_leave_alone)


prompt = "dts> "


class DTShell(Cmd):
    errors_loading = []

    # config = {}
    commands = {}
    core_commands = [
        "commands",
        "install",
        "uninstall",
        "update",
        "version",
        "exit",
        "help",
    ]

    shell_config: ShellConfig
    local_commands_info: CommandsInfo
    repo_info: RepoInfo
    commands_path: str

    include: types.SimpleNamespace

    def __init__(self, shell_config: ShellConfig, commands_info: CommandsInfo):
        self.shell_config = shell_config
        self.local_commands_info = commands_info

        self.intro = INTRO
        setattr(DTShell, "include", types.SimpleNamespace())

        # dtslogger.debug('sys.argv: %s' % sys.argv)
        check_if_outdated()

        self.repo_info = RepoInfo_for_version(shell_config.duckietown_version)
        self.commands_path = commands_path = self.local_commands_info.commands_path

        # add commands_path to the path of this session
        sys.path.insert(0, self.commands_path)
        # add third-party libraries dir to the path of this session
        sys.path.insert(0, os.path.join(self.commands_path, "lib"))

        # init commands
        cmds_just_initialized = False

        if exists(commands_path) and isfile(commands_path):
            remove(commands_path)
        if not exists(commands_path):
            msg = "I cannot find the command path %s" % commands_path
            if self.local_commands_info.leave_alone:
                raise Exception(msg)
            dtslogger.warning(msg)
            try:
                _init_commands(commands_path, self.repo_info)
            except InvalidRemote as e:
                msg = "I could not initialize the commands."
                raise CommandsLoadingException(msg) from e
            cmds_just_initialized = True

        # call super constructor
        super(DTShell, self).__init__()
        # remove the char `-` from the list of word separators, this allows us to suggest flags
        if self.use_rawinput and self.completekey:
            import readline

            readline.set_completer_delims(
                readline.get_completer_delims().replace("-", "", 1)
            )
        # check for updates (if needed)
        # Do not check it if we are using custom commands_path_leave_alone
        if (
            not cmds_just_initialized
            and not self.local_commands_info.leave_alone
            and not "update" in sys.argv
        ):
            check_commands_outdated(self.commands_path, self.repo_info)

        self.reload_commands()

    def save_config(self):
        write_shell_config(self.shell_config)

    def postcmd(self, stop, line):
        if len(line.strip()) > 0:
            print("")

    def emptyline(self):
        pass

    def complete(self, text, state):
        res = super(DTShell, self).complete(text, state)
        if res is not None:
            res += " "
        return res

    def reload_commands(self):
        # get installed commands
        installed_commands = self.commands.keys()
        for command in installed_commands:
            for a in ["do_", "complete_", "help_"]:
                if hasattr(DTShell, a + command):
                    delattr(DTShell, a + command)
        # re-install commands
        self.commands = _get_commands(self.commands_path)
        if self.commands is None:
            dtslogger.error("No commands found.")
            self.commands = {}
        # load commands
        # print('commands: %s' % self.commands)
        for cmd, subcmds in self.commands.items():
            # noinspection PyTypeChecker
            self._load_commands("", cmd, subcmds, 0)

        # TODO: load commands with prefix "challenges"

        if DTShell.errors_loading:
            msg = """


            !   Could not load commands.

                %s

            !   To recover, you might want to delete the directory
            !
            !      ~/.dt-shell/commands
            !
            !

            """ % "\n\n".join(
                DTShell.errors_loading
            )

            time.sleep(1)
            dtslogger.error(msg)
            time.sleep(5)

    def enable_command(self, command_name):
        if command_name in self.core_commands:
            return True
        # get list of all commands
        res = _get_commands(self.commands_path, all_commands=True)
        present = res.keys() if res is not None else []
        # enable if possible
        if command_name in present:
            flag_file = join(self.commands_path, command_name, "installed.user.flag")
            _touch(flag_file)
        return True

    def disable_command(self, command_name):
        if command_name in self.core_commands:
            return False
        # get list of all commands
        res = _get_commands(self.commands_path, all_commands=True)
        present = res.keys() if res is not None else []
        # enable if possible
        if command_name in present:
            flag_file = join(self.commands_path, command_name, "installed.user.flag")
            remove(flag_file)
        return True

    def _load_commands(
        self, package, command, sub_commands: Optional[Mapping[str, object]], lvl
    ):
        # load command
        klass = None
        error_loading = False
        if not sub_commands:
            spec = package + command + ".command.DTCommand"
            try:
                klass = _load_class(spec)
                # add loaded class to DTShell.include.<cmd_path>
                klass_path = [p for p in package.split(".") if len(p)]
                base = DTShell.include
                for p in klass_path:
                    if not hasattr(base, p):
                        setattr(base, p, types.SimpleNamespace())
                    base = getattr(base, p)
                setattr(base, command, klass)
            except UserError as e:
                raise
            except BaseException as e:
                # error_loading = True
                from .utils import format_exception

                se = format_exception(e)

                msg = "Cannot load command class %r (package=%r, command=%r): %s" % (
                    spec,
                    package,
                    command,
                    se,
                )
                # msg += ' sys.path: %s' % sys.path
                DTShell.errors_loading.append(msg)
                return

        # handle loading error and wrong class
        if error_loading:
            klass = DTCommandPlaceholder()
            if DEBUG:
                dtslogger.debug(
                    "ERROR while loading the command `%s`"
                    % (package + command + ".command.DTCommand",)
                )
        if not issubclass(klass.__class__, DTCommandAbs.__class__):
            klass = DTCommandPlaceholder()
            if DEBUG:
                dtslogger.debug(
                    "Command `%s` not found"
                    % (package + command + ".command.DTCommand",)
                )
        # initialize list of subcommands
        klass.name = command
        klass.level = lvl
        klass.commands = {}
        # attach first-level commands to the shell
        if lvl == 0:
            do_command = getattr(klass, "do_command")
            complete_command = getattr(klass, "complete_command")
            help_command = getattr(klass, "help_command")
            # wrap [klass, function] around a lambda function
            do_command_lam = lambda s, w: do_command(klass, s, w)
            complete_command_lam = lambda s, w, l, i, _: complete_command(
                klass, s, w, l, i, _
            )
            help_command_lam = lambda s: help_command(klass, s)
            # add functions do_* and complete_* to the shell
            setattr(DTShell, "do_" + command, do_command_lam)
            setattr(DTShell, "complete_" + command, complete_command_lam)
            setattr(DTShell, "help_" + command, help_command_lam)

        # stop recursion if there is no subcommand
        if sub_commands is None:
            return
        # load sub-commands
        for cmd, subcmds in sub_commands.items():
            if DEBUG:
                dtslogger.debug("DEBUG:: Loading %s" % package + command + ".*")
            # noinspection PyTypeChecker
            kl = self._load_commands(package + command + ".", cmd, subcmds, lvl + 1)
            if kl is not None:
                klass.commands[cmd] = kl
        # return class for this command
        return klass

    def get_dt1_token(self) -> str:
        if self.shell_config.token_dt1 is None:
            msg = 'Please set up a token for this using "dts tok set".'
            raise Exception(msg)
        else:
            return self.shell_config.token_dt1

    # noinspection PyMethodMayBeStatic
    def sprint(
        self, msg: str, color: Optional[str] = None, attrs: Sequence[str] = []
    ) -> None:
        return dts_print(msg=msg, color=color, attrs=attrs)

    def update_commands(self) -> bool:
        return update_commands(self.commands_path, self.repo_info)


def _touch(path: str) -> None:
    with open(path, "a"):
        utime(path, None)


def _load_class(name):
    if DEBUG:
        print("DEBUG:: Loading %s" % name)
    components = name.split(".")

    mod = __import__(components[0])

    for comp in components[1:]:
        try:
            mod = getattr(mod, comp)
        except AttributeError as e:
            msg = "Could not get field %r of module %r: %s" % (comp, mod.__name__, e)
            msg += "\n module file %s" % getattr(mod, "__file__", "?")
            raise AttributeError(msg)
    return mod



# -*- coding: utf-8 -*-
import json
import os
import subprocess
from datetime import datetime, timedelta
from typing import Optional, Tuple

import termcolor
import yaml
from whichcraft import which

from . import __version__, dtslogger
from .constants import DTShellConstants
from .exceptions import CouldNotGetVersion, NoCacheAvailable, URLException, UserError


def get_url(url, timeout=3):
    from six.moves import urllib

    try:
        req = urllib.request.Request(url)
        res = urllib.request.urlopen(req, timeout=timeout)
        content = res.read()
        if res.getcode() != 200:
            raise URLException(str(res))
        return content.decode("utf-8")
    except urllib.error.URLError:
        dtslogger.debug("Falling back to using curl because urllib failed.")
        if which("curl") is not None:
            cmd = ["curl", url, "-m", "2"]
            try:
                data = subprocess.check_output(cmd, stderr=subprocess.PIPE)
                return data
            except subprocess.CalledProcessError as e:
                msg = "Could not call %s: %s" % (cmd, e)
                raise URLException(msg)
        else:
            msg = "curl not available"
            raise URLException(msg)


def get_last_version_fresh() -> str:
    url = "https://pypi.org/pypi/duckietown-shell/json"

    try:
        try:
            data = get_url(url)
        except URLException as e:
            raise CouldNotGetVersion(str(e))
        try:
            info = json.loads(data)
        except BaseException as e:
            msg = "Could not read json %r" % data
            raise CouldNotGetVersion(msg) from e

        last_version = info["info"]["version"]
        return last_version
    except CouldNotGetVersion:
        raise
    except BaseException as e:
        raise CouldNotGetVersion() from e


def get_cache_filename() -> str:
    d0 = os.path.expanduser(DTShellConstants.ROOT)
    return os.path.join(d0, "pypi-cache.yaml")


def read_cache() -> Tuple[str, datetime]:
    try:
        fn = get_cache_filename()
        if os.path.exists(fn):
            data = open(fn).read()
            interpreted = yaml.load(data, Loader=yaml.Loader)
            version = interpreted["version"]
            dt = interpreted["timestamp"]
            return version, dt
        else:
            msg = "File %s does not exist." % fn
            raise NoCacheAvailable(msg)
    except Exception as e:
        msg = "Could not read cache: %s" % e
        raise NoCacheAvailable(msg)


def write_cache(version: str, dt: datetime) -> None:
    fn = get_cache_filename()
    d0 = os.path.dirname(fn)
    if not os.path.exists(d0):
        os.makedirs(d0)

    with open(fn, "w") as f:
        contents = dict(version=version, timestamp=dt)
        y = yaml.dump(contents)
        f.write(y)


def get_last_version() -> Optional[str]:
    now = datetime.now()
    update = False

    version = None
    try:
        version, timestamp = read_cache()
    except NoCacheAvailable:
        timestamp = now
        update = True

    if not update:
        delta = now - timestamp
        if delta > timedelta(minutes=10):
            # dtslogger.debug('Version cache is outdated (%s).' % delta)
            update = True

    if update:
        # dtslogger.debug('Getting last version from PyPI.')
        try:
            version = get_last_version_fresh()
            write_cache(version, now)
            return version
        except CouldNotGetVersion:
            return None

    # XXX: this might not be set
    return version


def is_older(a: str, b: str) -> bool:
    na = tuple(map(int, a.split(".")))
    nb = tuple(map(int, b.split(".")))

    return na < nb


def check_if_outdated() -> None:
    latest_version = get_last_version()
    # print('last version: %r' % latest_version)
    # print('installed: %r' % __version__)
    if latest_version and is_older(__version__, latest_version):
        msg = """

There is an updated duckietown-shell available.

  You have: {current}

 available: {available} 

You must update the shell using `pip`.        
        
        """.format(
            current=__version__, available=latest_version
        )
        print(termcolor.colored(msg, "yellow"))
        raise UserError(msg)
        # wait = 3
        # time.sleep(1)
        # print('Waiting %d seconds to give you time to read the message.' % wait)
        # time.sleep(wait)
        # return True
    # return False



import logging

__all__ = []


def setup_logging_format():
    from logging import Logger, StreamHandler, Formatter
    import logging

    FORMAT = "%(name)15s|%(filename)15s:%(lineno)-4s - %(funcName)-15s| %(message)s"

    logging.basicConfig(format=FORMAT)

    # noinspection PyUnresolvedReferences
    root = Logger.root
    if root.handlers:
        for handler in root.handlers:
            if isinstance(handler, StreamHandler):
                formatter = Formatter(FORMAT)
                handler.setFormatter(formatter)
    else:
        logging.basicConfig(format=FORMAT)


def add_coloring_to_emit_ansi(fn):
    # add methods we need to the class
    def new(*args):
        levelno = args[1].levelno
        if levelno >= 50:
            color = "\x1b[31m"  # red
        elif levelno >= 40:
            color = "\x1b[31m"  # red
        elif levelno >= 30:
            color = "\x1b[33m"  # yellow
        elif levelno >= 20:
            color = "\x1b[32m"  # green
        elif levelno >= 10:
            color = "\x1b[35m"  # pink
        else:
            color = "\x1b[0m"  # normal

        msg = str(args[1].msg)

        lines = msg.split("\n")

        def color_line(l):
            return "%s%s%s" % (color, l, "\x1b[0m")  # normal

        lines = list(map(color_line, lines))

        args[1].msg = "\n".join(lines)
        return fn(*args)

    return new


def setup_logging_color():
    import platform

    if platform.system() != "Windows":
        emit2 = add_coloring_to_emit_ansi(logging.StreamHandler.emit)
        logging.StreamHandler.emit = emit2


def setup_logging():
    # logging.basicConfig()
    setup_logging_color()
    setup_logging_format()



import traceback
from typing import Optional

import termcolor


def indent(s: str, prefix: str, first: Optional[str] = None) -> str:
    s = str(s)
    assert isinstance(prefix, str)
    lines = s.split("\n")
    if not lines:
        return ""

    if first is None:
        first = prefix

    m = max(len(prefix), len(first))

    prefix = " " * (m - len(prefix)) + prefix
    first = " " * (m - len(first)) + first

    # different first prefix
    res = ["%s%s" % (prefix, line.rstrip()) for line in lines]
    res[0] = "%s%s" % (first, lines[0].rstrip())
    return "\n".join(res)


def raise_wrapped(etype, e, msg, compact=False, exc=None, **kwargs):
    """ Raises an exception of type etype by wrapping
        another exception "e" with its backtrace and adding
        the objects in kwargs as formatted by format_obs.

        if compact = False, write the whole traceback, otherwise just str(e).

        exc = output of sys.exc_info()
    """

    e = raise_wrapped_make(etype, e, msg, compact=compact, **kwargs)

    #     if exc is not None:
    #         _, _, trace = exc
    #         raise etype, e.args, trace
    #     else:
    raise e


def raise_wrapped_make(etype, e, msg, compact=False, **kwargs):
    """ Constructs the exception to be thrown by raise_wrapped() """
    assert isinstance(e, BaseException), type(e)
    assert isinstance(msg, str), type(msg)
    s = msg

    import sys

    if sys.version_info[0] >= 3:
        es = str(e)
    else:
        if compact:
            es = str(e)
        else:
            es = format_exception(e)

    s += "\n" + indent(es.strip(), "| ")

    return etype(s)


def check_isinstance(ob, expected, **kwargs):
    if not isinstance(ob, expected):
        kwargs["object"] = ob
        raise_type_mismatch(ob, expected, **kwargs)


def raise_type_mismatch(ob, expected, **kwargs):
    """ Raises an exception concerning ob having the wrong type. """
    e = "Object not of expected type:"
    e += "\n  expected: %s" % str(expected)
    e += "\n  obtained: %s" % str(type(ob))
    # e += '\n' + indent(format_obs(kwargs), ' ')
    raise ValueError(e)


SPACE_TAG = "SPACE_TAG"


def replace_spaces(x: str) -> str:
    return x.replace(" ", SPACE_TAG)


def undo_replace_spaces(x: str) -> str:
    return x.replace(SPACE_TAG, " ")


def format_exception(e):
    return traceback.format_exc()  # None, e)


def href(x):
    return termcolor.colored(x, "blue", attrs=["underline"])


def dark_yellow(x):
    return termcolor.colored(x, "yellow", attrs=[])


def dark(x):
    return termcolor.colored(x, attrs=["dark"])
